//#####################################################################
// Copyright 2004-2008, Zhaosheng Bao, Eilene Hao, Sergey Koltakov, Frank Losasso, Craig Schroeder, Andrew Selle, Michael Turitzin.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Log/DEBUG_UTILITIES.h>
#include <PhysBAM_Tools/Parsing/PARAMETER_LIST.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Lights/RENDERING_DIRECTIONAL_LIGHT.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Lights/RENDERING_POINT_LIGHT.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Lights/RENDERING_SPOTLIGHT.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Shaders/RENDERING_LIGHT_SHADER.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering/RENDER_WORLD.h>
#include "GENERIC_RENDER_EXAMPLE.h"
using namespace PhysBAM;
//#####################################################################
// Light - Prepare lights
//#####################################################################
template<class T,class RW> void GENERIC_RENDER_EXAMPLE<T,RW>::
Light(RENDER_WORLD<T>& world,const int frame,PARAMETER_LIST& parameters)
{
    std::string type=parameters.Get_Parameter("Type",std::string("Null"));
    bool global_photons=parameters.Get_Parameter("Supports_Global_Photons",true);bool caustic_photons=parameters.Get_Parameter("Supports_Caustic_Photons",true);
    bool volume_photons=parameters.Get_Parameter("Supports_Volume_Photons",true);
    bool photon_source_only=parameters.Get_Parameter("Photon_Source_Only",false);
    bool casts_shadows=parameters.Get_Parameter("Casts_Shadows",true);
    if(type=="Point"){
        TV position=current_transform.Homogeneous_Times(parameters.Get_Parameter("Position",TV()));
        TV color=current_transform.Homogeneous_Times(parameters.Get_Parameter("Color",TV(1,1,1)));
        T power=parameters.Get_Parameter("Power",(T)1);
        lights.push_back(new RENDERING_POINT_LIGHT<T>(position,color,power,true,world,global_photons,caustic_photons,volume_photons));
        LOG::cout<<"Light  Point position="<<position<<" color="<<color<<" power="<<power<<std::endl;}
    else if(type=="Spotlight"){
        TV position=current_transform.Homogeneous_Times(parameters.Get_Parameter("Position",TV()));
        TV color=current_transform.Homogeneous_Times(parameters.Get_Parameter("Color",TV(1,1,1)));
        T power=parameters.Get_Parameter("Power",(T)1);
        TV direction=current_transform.Transposed().Inverse().Homogeneous_Times(parameters.Get_Parameter("Direction",TV(0,0,1)));
        T cone_angle=parameters.Get_Parameter("Cone_Angle",(T)pi/4);
        T penumbra_angle=parameters.Get_Parameter("Penumbra_Angle",(T)pi/4);
        lights.push_back(new RENDERING_SPOTLIGHT<T>(position,color,power,direction,cone_angle,penumbra_angle,world,global_photons,caustic_photons,volume_photons,photon_source_only));}
    else if(type=="Directional") {
        TV direction=current_transform.Homogeneous_Times(parameters.Get_Parameter("Direction",TV()));
        TV color=current_transform.Homogeneous_Times(parameters.Get_Parameter("Color",TV(1,1,1)));
        T power=parameters.Get_Parameter("Power",(T)1);
        lights.push_back(new RENDERING_DIRECTIONAL_LIGHT<T>(direction,color,power,world));
        LOG::cout<<"Light  Directional direction="<<direction<<" color="<<color<<" power="<<power<<std::endl;}
    else PHYSBAM_FATAL_ERROR("Unknown light type:"+type);
    lights.back()->casts_shadows=casts_shadows;
}
//#####################################################################
template void GENERIC_RENDER_EXAMPLE<float,float>::Light(RENDER_WORLD<float>& world,const int frame,PARAMETER_LIST& parameters);
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void GENERIC_RENDER_EXAMPLE<double,float>::Light(RENDER_WORLD<double>& world,const int frame,PARAMETER_LIST& parameters);
template void GENERIC_RENDER_EXAMPLE<double,double>::Light(RENDER_WORLD<double>& world,const int frame,PARAMETER_LIST& parameters);
#endif
