//#####################################################################
// Copyright 2004-2008, Zhaosheng Bao, Eilene Hao, Jeong-Mo Hong, Geoffrey Irving, Sergey Koltakov, Frank Losasso, Avi Robinson-Mosher, Andrew Selle, Tamar Shinar, Jerry Talton, Michael Turitzin.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Grids_RLE_Interpolation/AVERAGING_RLE.h>
#include <PhysBAM_Tools/Grids_Uniform_Computations/GRADIENT_UNIFORM.h>
#include <PhysBAM_Tools/Parsing/PARAMETER_LIST.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Point_Clouds/READ_WRITE_POINT_CLOUD.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Lights/RENDERING_LIGHT.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_BOX.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_CYLINDER.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_PARTICLES.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_PLANE.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_SPHERE.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_TRIANGLE.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_UNIFORM_VOXELS.h>
#include <PhysBAM_Rendering/PhysBAM_Ray_Tracing/Rendering_Objects/RENDERING_WALL.h>
#include "GENERIC_RENDER_EXAMPLE.h"
using namespace PhysBAM;
//#####################################################################
// Object
//#####################################################################
template<class T,class RW> void GENERIC_RENDER_EXAMPLE<T,RW>::
Object(RENDER_WORLD<T>& world,const int frame,PARAMETER_LIST& parameters)
{
    std::string type=parameters.Get_Parameter("Type",std::string("Null"));
    std::string name=parameters.Get_Parameter("Name",std::string("Name"));
    int priority=parameters.Get_Parameter("Priority",0);
    bool support_transparent_overlapping_objects=parameters.Get_Parameter("Support_Transparent_Overlapping_Objects",false);
    LOG::cout<<"Object has priority "<<priority<<" and support for overlapping transparent objects is: "<< support_transparent_overlapping_objects<<std::endl;
    RENDERING_OBJECT<T> *object=0;
    ARRAY<RENDERING_OBJECT<T>*> object_group;
    LOG::cout<<"---=-=-=-=Parsing object..."<<std::endl;
    if(type=="Sphere"){
        TV position=parameters.Get_Parameter("Position",TV(0,0,0));
        T radius=parameters.Get_Parameter("Radius",(T)1);
        object=new RENDERING_SPHERE<T>(position,radius);object->add_to_spatial_partition=true;
        LOG::cout<<"Object '"<<name<<"'Sphere radius="<<radius<<" position="<<position<<std::endl;}
    else if(type=="Plane"){
        LOG::cout<<"Plane"<<std::endl;
        TV position=parameters.Get_Parameter("Position",TV(0,0,0));
        TV normal=parameters.Get_Parameter("Normal",TV(0,1,0));
        RENDERING_PLANE<T>* plane=new RENDERING_PLANE<T>(normal,position);
        object=plane;
        plane->texture_vector1=parameters.Get_Parameter("Texture_Vector1",TV(1,0,0));
        plane->texture_vector2=parameters.Get_Parameter("Texture_Vector2",TV(0,0,1));}
    else if(type=="Box"){
        LOG::cout<<"Box"<<std::endl;
        T xmin=parameters.Get_Parameter("Xmin",(T)-1);T xmax=parameters.Get_Parameter("Xmax",(T)1);
        T ymin=parameters.Get_Parameter("Ymin",(T)-1);T ymax=parameters.Get_Parameter("Ymax",(T)1);
        T zmin=parameters.Get_Parameter("Zmin",(T)-1);T zmax=parameters.Get_Parameter("Zmax",(T)1);
        object=new RENDERING_BOX<T>(xmin,xmax,ymin,ymax,zmin,zmax);}
     else if(type=="Cylinder"){
        TV X1=parameters.Get_Parameter("X1",TV(0,-1,0));
        TV X2=parameters.Get_Parameter("X2",TV(0,1,0));
        T radius=parameters.Get_Parameter("Radius",(T)1);
        LOG::cout<<"Cylinder principal points X1="<<X1<<" X2="<<X2<<" radius="<<radius<<std::endl;
        object=new RENDERING_CYLINDER<T>(X1,X2,radius);}
    else if(type=="Triangle"){
        LOG::cout<<"Triangle"<<std::endl;
        TV x1=parameters.Get_Parameter("X1",TV(1,0,0));
        TV x2=parameters.Get_Parameter("X2",TV(0,1,0));
        TV x3=parameters.Get_Parameter("X3",TV(0,0,1));
        object=new RENDERING_TRIANGLE<T>(x1,x2,x3);}
    else if(type=="Wall"){
        LOG::cout<<"Wall"<<std::endl;
        std::string control=parameters.Get_Parameter("Control",std::string("Null"));
        bool xmin_wall=parameters.Get_Parameter("Show_Xmin", true); bool xmax_wall=parameters.Get_Parameter("Show_Xmax",true);
        bool ymin_wall=parameters.Get_Parameter("Show_Ymin", true); bool ymax_wall=parameters.Get_Parameter("Show_Ymax",true);
        bool zmin_wall=parameters.Get_Parameter("Show_Zmin", true); bool zmax_wall=parameters.Get_Parameter("Show_Zmax",true);
        bool just_frame=parameters.Get_Parameter("Just_Frame", false);
        BOX<TV>* box=0;
        if (control=="Null"||control=="Free"){
            T xmin=parameters.Get_Parameter("Xmin",(T)-1);T xmax=parameters.Get_Parameter("Xmax",(T)1);
            T ymin=parameters.Get_Parameter("Ymin",(T)-1);T ymax=parameters.Get_Parameter("Ymax",(T)1);
            T zmin=parameters.Get_Parameter("Zmin",(T)-1);T zmax=parameters.Get_Parameter("Zmax",(T)1);
            box=new BOX<TV>(xmin,xmax,ymin,ymax,zmin,zmax);}
        else if(control=="Object"){ // any object has a bounding box so just use it...
            std::string object_name=parameters.Get_Parameter("Object",std::string("Null"));
            RENDERING_OBJECT<T>* object=0;
            if(!objects.Get(object_name,object)){LOG::cout<<"Unknown rendering object '"<<object_name<<"'specified for Wall"<<std::endl;exit(1);}
            box=new BOX<TV>(object->World_Space_Bounding_Box());}
        T shrink=parameters.Get_Parameter("Shrink", (T)0);
        box->Change_Size(-shrink);
        T thickness=parameters.Get_Parameter("Thickness",(T)0);
        if(thickness){
            // thick walls are just a group of rendering boxes
            BOX<TV> outer_box=*box;
            T perturb_value=(T)0.0005;
            if(xmin_wall){
                outer_box.min_corner.x=box->min_corner.x-thickness;outer_box.max_corner.x=box->min_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z-thickness;outer_box.max_corner.z=box->min_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->min_corner.x-thickness;outer_box.max_corner.x=box->min_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->max_corner.z-thickness;outer_box.max_corner.z=box->max_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->min_corner.x-thickness;outer_box.max_corner.x=box->min_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->min_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->min_corner.x-thickness;outer_box.max_corner.x=box->min_corner.x+thickness;
                outer_box.min_corner.y=box->max_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                if(!just_frame){
                  outer_box.min_corner.x=box->min_corner.x-thickness;outer_box.max_corner.x=box->min_corner.x+thickness;
                  outer_box.min_corner.y=box->min_corner.y+thickness-perturb_value;outer_box.max_corner.y=box->max_corner.y-thickness+perturb_value;
                  outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
            if(xmax_wall){
                outer_box.min_corner.x=box->max_corner.x-thickness;outer_box.max_corner.x=box->max_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z-thickness;outer_box.max_corner.z=box->min_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->max_corner.x-thickness;outer_box.max_corner.x=box->max_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->max_corner.z-thickness;outer_box.max_corner.z=box->max_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->max_corner.x-thickness;outer_box.max_corner.x=box->max_corner.x+thickness;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->min_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->max_corner.x-thickness;outer_box.max_corner.x=box->max_corner.x+thickness;
                outer_box.min_corner.y=box->max_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                if(!just_frame){
                  outer_box.min_corner.x=box->max_corner.x-thickness;outer_box.max_corner.x=box->max_corner.x+thickness;
                  outer_box.min_corner.y=box->min_corner.y+thickness-perturb_value;outer_box.max_corner.y=box->max_corner.y-thickness+perturb_value;
                  outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
            if(ymin_wall){
                if(!just_frame){
                  outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                  outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->min_corner.y+thickness;
                  outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
            if(ymax_wall){
                if(!just_frame){
                  outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                  outer_box.min_corner.y=box->max_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                  outer_box.min_corner.z=box->min_corner.z+thickness-perturb_value;outer_box.max_corner.z=box->max_corner.z-thickness+perturb_value;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
            if(zmin_wall){
                outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                outer_box.min_corner.y=box->max_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z-thickness;outer_box.max_corner.z=box->min_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->min_corner.y+thickness;
                outer_box.min_corner.z=box->min_corner.z-thickness;outer_box.max_corner.z=box->min_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                if(!just_frame){
                  outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                  outer_box.min_corner.y=box->min_corner.y+thickness-perturb_value;outer_box.max_corner.y=box->max_corner.y-thickness+perturb_value;
                  outer_box.min_corner.z=box->min_corner.z-thickness;outer_box.max_corner.z=box->min_corner.z+thickness;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
            if(zmax_wall){
                outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                outer_box.min_corner.y=box->max_corner.y-thickness;outer_box.max_corner.y=box->max_corner.y+thickness;
                outer_box.min_corner.z=box->max_corner.z-thickness;outer_box.max_corner.z=box->max_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                outer_box.min_corner.y=box->min_corner.y-thickness;outer_box.max_corner.y=box->min_corner.y+thickness;
                outer_box.min_corner.z=box->max_corner.z-thickness;outer_box.max_corner.z=box->max_corner.z+thickness;
                object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));

                if(!just_frame){
                  outer_box.min_corner.x=box->min_corner.x+thickness-perturb_value;outer_box.max_corner.x=box->max_corner.x-thickness+perturb_value;
                  outer_box.min_corner.y=box->min_corner.y+thickness-perturb_value;outer_box.max_corner.y=box->max_corner.y-thickness+perturb_value;
                  outer_box.min_corner.z=box->max_corner.z-thickness;outer_box.max_corner.z=box->max_corner.z+thickness;
                  object_group.Append(new RENDERING_BOX<T>(outer_box.min_corner.x,outer_box.max_corner.x,outer_box.min_corner.y,outer_box.max_corner.y,outer_box.min_corner.z,outer_box.max_corner.z));
                }
            }
        }
        else{
            T texture_scale=parameters.Get_Parameter("Texture_Scale", (T)1);
            object=new RENDERING_WALL<T>(*box, xmin_wall, xmax_wall, ymin_wall, ymax_wall, zmin_wall, zmax_wall,texture_scale);}}
    else if(type=="Voxel_Data"){
        LOG::cout<<"Voxel"<<std::endl;
        bool precompute=parameters.Get_Parameter("Precompute_Single_Scattering",false);
        std::string grid_filename=parameters.Get_Parameter("Grid_Filename",std::string("<unknown>"));
        std::string coarse_grid_filename=parameters.Get_Parameter("Coarse_Grid_Filename",std::string("<unknown>"));
        std::string density_filename=parameters.Get_Parameter("Density_Filename",std::string("<unknown>"));
        std::string density_fraction_filename=parameters.Get_Parameter("Density_Fraction_Filename",std::string("<unknown>"));
        std::string temperature_filename=parameters.Get_Parameter("Temperature_Filename",std::string("<unknown>"));

        bool use_density_gradient=parameters.Get_Parameter("Use_Density_Gradient",false);
        bool use_cubic_for_density=parameters.Get_Parameter("Use_Density_Cubic",false);
        T density_scale=parameters.Get_Parameter("Density_Scale",(T)1);
        T density_offset=parameters.Get_Parameter("Density_Offset",(T)0);
        bool clamp_low_density=parameters.Get_Parameter("Clamp_Low_Density",false);
        T density_lowest=parameters.Get_Parameter("Density_Lowest",(T)0);
        bool clamp_high_density=parameters.Get_Parameter("Clamp_High_Density",false);
        T density_highest=parameters.Get_Parameter("Density_Highest",(T)0);

        T temperature_scale=parameters.Get_Parameter("Temperature_Scale",(T)1);
        T temperature_offset=parameters.Get_Parameter("Temperature_Offset",(T)0);
        bool clamp_low_temperature=parameters.Get_Parameter("Clamp_Low_Temperature",false);
        T temperature_lowest=parameters.Get_Parameter("Temperature_Lowest",(T)0);
        bool clamp_high_temperature=parameters.Get_Parameter("Clamp_High_Temperature",false);
        T temperature_highest=parameters.Get_Parameter("Temperature_Highest",(T)0);
        
        T volume_step=parameters.Get_Parameter("Volume_Step",(T)0.1);
        // always read densities but sometimes read temperatures
        GRID<TV>* grid=new GRID<TV>;ARRAY<T,VECTOR<int,3> >* density_data=new ARRAY<T,VECTOR<int,3> >;
        FILE_UTILITIES::template Read_From_File<RW>(Animated_Filename(grid_filename,frame),*grid);
        FILE_UTILITIES::template Read_From_File<RW>(Animated_Filename(density_filename,frame),*density_data);
        if(use_density_gradient){
            LOG::cout<<"Using density gradient"<<std::endl;
            ARRAY<T,VECTOR<int,3> >* density_gradient_data=new ARRAY<T,VECTOR<int,3> >(*density_data);
            GRADIENT::Compute_Magnitude(*grid,0,*density_data,*density_gradient_data);
            delete density_data;
            density_data=density_gradient_data;}
        if(density_fraction_filename!="<unknown>"){
            LOG::cout<<"Using density fraction from "<<density_fraction_filename<<std::endl;
            ARRAY<T,VECTOR<int,3> > density_fraction_data;
            FILE_UTILITIES::template Read_From_File<RW>(Animated_Filename(density_fraction_filename,frame),density_fraction_data);
            *density_data*=density_fraction_data;}
        if(use_cubic_for_density){
            LOG::cout<<"Using density cubic"<<std::endl;
            for(typename GRID<TV>::CELL_ITERATOR iterator(*grid);iterator.Valid();iterator.Next()){VECTOR<int,TV::dimension> cell=iterator.Cell_Index();(*density_data)(cell)=(*density_data)(cell)*(*density_data)(cell)*(*density_data)(cell);}}
        LOG::cout<<"  Read grid " <<*grid;
        RENDERING_UNIFORM_VOXELS<T> *voxels;
        if(coarse_grid_filename!="<unknown>"){
            GRID<TV>* coarse_grid=new GRID<TV>;FILE_UTILITIES::template Read_From_File<RW>(Animated_Filename(coarse_grid_filename,frame),*coarse_grid);
            voxels=new RENDERING_UNIFORM_VOXELS<T>(*grid,*coarse_grid,*density_data,(T)volume_step);}
        else voxels=new RENDERING_UNIFORM_VOXELS<T>(*grid,*density_data,(T)volume_step);
        if(temperature_filename!="<unknown>"){
            ARRAY<T,VECTOR<int,3> >* temperature_data=new ARRAY<T,VECTOR<int,3> >;voxels->data.Append(temperature_data);
            FILE_UTILITIES::template Read_From_File<RW>(Animated_Filename(temperature_filename,frame),*temperature_data);}

        voxels->data_scale.Append(density_scale);
        voxels->data_scale.Append(temperature_scale);

        voxels->data_offset.Append(density_offset);
        voxels->data_offset.Append(temperature_offset);

        voxels->data_clamp_low_value.Append(clamp_low_density);
        voxels->data_clamp_low_value.Append(clamp_low_temperature);

        voxels->data_lowest_value.Append(density_lowest);
        voxels->data_lowest_value.Append(temperature_lowest);

        voxels->data_clamp_high_value.Append(clamp_high_density);
        voxels->data_clamp_high_value.Append(clamp_high_temperature);

        voxels->data_highest_value.Append(density_highest);
        voxels->data_highest_value.Append(temperature_highest);

        voxels->Print_Parameters();
        
        voxels->number_of_smoothing_steps=parameters.Get_Parameter("Number_Of_Smoothing_Steps",(int)3);
        voxels->precompute_single_scattering=precompute;
        object=voxels;}
    else PHYSBAM_FATAL_ERROR("Unknown object type "+type);
    // Register the object if it was created
    if(object){    // shaders of levelset-multiple are already assigned to region objects.
        std::string shader_name=parameters.Get_Parameter("Shader",std::string("<unknown>"));
        std::string volume_shader_name=parameters.Get_Parameter("Volume_Shader",std::string("<unknown>"));
        object->index_of_refraction=parameters.Get_Parameter("Index_Of_Refraction",(T)1);
        // check to make sure at least one shader specified
        if(shader_name=="<unknown>"&&volume_shader_name=="<unknown>") PHYSBAM_FATAL_ERROR("No shader (volume or surface) specified for object "+name);
        // surface shader
        if(shader_name!="<unknown>"){
            if(!shaders.Get(shader_name,object->material_shader)) PHYSBAM_FATAL_ERROR(STRING_UTILITIES::string_sprintf("Invalid shader '%s' specified for object %s\n",shader_name.c_str(),name.c_str()));}
        // volume shader
        if(volume_shader_name!="<unknown>"){
            if(!volume_shaders.Get(volume_shader_name,object->volumetric_shader))
                PHYSBAM_FATAL_ERROR(STRING_UTILITIES::string_sprintf("Invalid volume shader '%s' specified for object %s\n",volume_shader_name.c_str(),name.c_str()));
            object->add_to_spatial_partition=false;}
        object->name=name;object->Update_Transform(current_transform);objects.Set(name,object);
        object->priority=priority;object->support_transparent_overlapping_objects=support_transparent_overlapping_objects;}
    else if(object_group.m){
        std::string shader_name=parameters.Get_Parameter("Shader",std::string("<unknown>"));
        std::string volume_shader_name=parameters.Get_Parameter("Volume_Shader",std::string("<unknown>"));
        T index_of_refraction=parameters.Get_Parameter("Index_Of_Refraction",(T)1);
        for(int i=1;i<=object_group.m;i++){
            object_group(i)->index_of_refraction=index_of_refraction;
            object_group(i)->priority=priority;object_group(i)->support_transparent_overlapping_objects=support_transparent_overlapping_objects;}
        // check to make sure at least one shader specified
        if(shader_name=="<unknown>"&&volume_shader_name=="<unknown>") PHYSBAM_FATAL_ERROR(STRING_UTILITIES::string_sprintf("No shader (volume or surface) specified for object %s\n",name.c_str()));
        // surface shader
        if(shader_name!="<unknown>"){
            MATERIAL_SHADER<T>* material_shader=0;
            if(!shaders.Get(shader_name,material_shader)) PHYSBAM_FATAL_ERROR(STRING_UTILITIES::string_sprintf("Invalid shader '%s' specified for object %s\n",shader_name.c_str(),name.c_str()));
            for(int i=1;i<=object_group.m;i++) object_group(i)->material_shader=material_shader;}
        // volume shader
        if(volume_shader_name!="<unknown>"){
            VOLUMETRIC_SHADER<T>* volumetric_shader=0;
            if(!volume_shaders.Get(volume_shader_name,volumetric_shader)) PHYSBAM_FATAL_ERROR(STRING_UTILITIES::string_sprintf("Invalid volume shader '%s' specified for object %s\n",volume_shader_name.c_str(),name.c_str()));
            for(int i=1;i<=object_group.m;i++){object_group(i)->volumetric_shader=volumetric_shader;object_group(i)->add_to_spatial_partition=false;}}
        for(int i=1;i<=object_group.m;i++){
            std::string instance_name=STRING_UTILITIES::string_sprintf("%s_%d",name.c_str(),i);
            object_group(i)->name=instance_name;object_group(i)->Update_Transform(current_transform);objects.Set(instance_name,object_group(i));}}
    if(object) GENERIC_RENDER_EXAMPLE<T,RW>::Add_Solid_Texture(object,parameters);
}
//#####################################################################
template void GENERIC_RENDER_EXAMPLE<float,float>::Object(RENDER_WORLD<float>& world,const int frame,PARAMETER_LIST& parameters);
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void GENERIC_RENDER_EXAMPLE<double,float>::Object(RENDER_WORLD<double>& world,const int frame,PARAMETER_LIST& parameters);
template void GENERIC_RENDER_EXAMPLE<double,double>::Object(RENDER_WORLD<double>& world,const int frame,PARAMETER_LIST& parameters);
#endif
